function [T, X] = bacterial_fitness_decr_eff_sim(deb,h,fin,theta,theta_fixe,y0,coeff_decr,std,control)
%DESCRIPTION: %Simulate data for bacterial population under treatment with decreasing of antibiotic defficiency over time
%USAGE:  [T, X] = bacterial_fitness_decr_eff_sim(deb,h,fin,theta,theta_fixe,y0,coeff_decr,std,control)
%INPUT :
%deb              :(real) beginning of the observation interval
%h                :(real) integration step
%fin              :(real) end of the observation interval
%theta            :(vector) vector of parameters
%theta_fixe       :(vector) complementary vector of parameters
%y0               :(vector) initial conditions for the system
%coeff_decr       :(positive real) efficency loss coefficient (parameter c in the mu_s and m_r functions)
%std              :(positive real) variance level of measurement noise
%OPTIONAL
%control          :(function handler) exogenous forcing function (default zero functions)
%OUTPUT:
%T                    :(vector) vector of observation time 
%X                    :(matrix or matlab solution object) simulated model values OR matlab solution object if nargout == 1         
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 

pres_control = 0;
if nargin > 8
    pres_control= 1;
end

temp = [deb :h :fin];
T = temp;

gamma = theta(1);
rho = theta(2);
psi = theta(3);
c = theta(4);



beta = theta_fixe(1);
Pmax = theta_fixe(2);
delta = theta_fixe(3);
delta_p = theta_fixe(4);
nu_s = theta_fixe(5);
K = theta_fixe(6);
E_s =  theta_fixe(7);
A = theta_fixe(8);
A_s50 = theta_fixe(9);
gamma_max = theta_fixe(10);
A_s = theta_fixe(11);
gamma_50 = theta_fixe(12);
lambda = theta_fixe(13);
A_r50 = theta_fixe(14);
E_r = theta_fixe(15); 


zeta = gamma_max*A_s/(A_s+gamma_50);


function dy = equ(t,y)
    
    mu_s = E_s*A/((A+A_s50)*(1+coeff_decr*t/fin));
    mu_r = E_r*A/((A+A_r50)*(1+coeff_decr*t/fin));

    dy = zeros(3,1);
    dy(1) = beta*(y(2)+y(3))*(1-y(1)/Pmax) - delta*(y(2)+y(3))*y(1)-delta_p*y(1);
	dy(2) = nu_s*y(2)*(1-(y(2)+y(3))/K)-mu_s*y(2)-(gamma+zeta)*y(1)*y(2)-lambda*y(2)*y(3)+rho*y(3)-psi*y(2);
    dy(3) = (1-c)*nu_s*y(3)*(1-(y(2)+y(3))/K)- mu_r*y(3)-(gamma+zeta)*y(1)*y(3)+lambda*y(2)*y(3)-rho*y(3)-psi*y(3);

	if pres_control == 1
		dy = dy + control(t);
	end

end

if nargout > 1 
[T,Y] = ode45(@equ,temp,y0);

std_vect = h*sum(Y)*std/((fin-deb)*100);
res = Y + repmat(std_vect,length(temp),1).*randn(length(temp),3);

X =  res;

else
    sol = ode45(@equ,temp,y0);
    T = sol;
end

%plot(T,Y(:,1),'b',T,res(:,1),'bo');
%plot(T,Y(:,1),'b',T,res(:,1),'bo',T,Y(:,2),'r',T,res(:,2),'rx');
%plot(T,Y(:,1),'b',T,Y(:,2),'r');
%plot(T,res(:,1),'bx',T,res(:,2),'rx');

end