function [est_param,fval] = est_param_pontryagin_bvp4c(bs,vect_field,der_vect_field,der_vect_field_theta,der_vect_field_state_theta,der_vect_field_state_state,mat_Obs,mat_U,param_ini,x_0,param_fixe)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%DESCRIPTION: % Plot the economic function corresponding to the pontryagin estimator for non linear or linear model
%
%USAGE:   res_fval = plot_func_eco_pontryagin_general_case(bs,vect_field,der_vect_field,der_sec_vect_field,mat_Obs,mat_U,param_trial,x_0,param_fixe,type_solver)
%
%INPUT :
%bs               :(bSplines  object) , embed the nonparametric estimators
%vect_field       :(function handler) embed the function corresponding to the vector field with signature: 
%                  res_t =  vect_field(t,vx,param,param_fixe)
%                           t : time
%                           vx : state value at time t
%                           param : varying parameter value for the plotted graph
%                           param_fixe : other parameter value
%                           res_t : (vector of size dim x 1) vector field  value at time t
%der_vect_field       :(function handler) embed the function corresponding to the vector field derivative form x.rt state with signature: 
%                  res_t = der_vect_field(t,vx,param,param_fixe)
%                           t : time
%                           vx : state value at time t
%                           param : varying parameter value for the plotted graph
%                           param_fixe : other parameter value
%                           res_t : (matrix of size dim x dim) matrix der_f_i/x_j: vector field derivative form x.rt state value at time t
%der_sec_vect_field       :(function handler) embed the function corresponding to the vector field second derivative form x.rt state with signature: 
%                  res_t = der_sec_vect_field(t,vx,param,param_fixe)
%                           t : time
%                           vx : state value at time t
%                           param : varying parameter value for the plotted graph
%                           param_fixe : other parameter value
%                           res_t : (hypermatrix of size dim x dim x dim) matrix der_sec_f_i/x_j_x_h: vector field second derivative form x.rt state value at time t
%mat_Obs                  : (matrix of size dim_obs x dim): observation matrix C
%mat_U                    : (matrix of size dim x dim): ponderation matrix U for control
%param_trial              : (matrix of size 1 x 2 or 2 x 2) lower and upper bound for varying parameter(s) value
%OPTIONAL
%x_0                      :(vector of size 1 x dim) initial condition if known (default empty ([]) )
%param_fixe               :(vector) fixed parameter for the model (default empty ([]) )
%type_solver              : (integer) indicate the used mixed boundary ode solver
%                         0 (default) extremal variation algorithm
%                         1           quasilinearisation algorithm
%
%OUTPUT:
%res_fval                 :(matrix or vector) economical function value for each varying parameter(s) value
%               
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 

    Obs = bs.getObservation();
    T_c = Obs.getTimes();
    Times = T_c{1};
    invU = inv(mat_U);
    dim_syst = length(mat_Obs(1,:));
    dim_obs = length(mat_Obs(:,1));

    pas_solve =200;
    Int_ode_solve = Times(1):((Times(end)-Times(1))/pas_solve):Times(end);          
    
    free_ci = 0;
    if nargin < 10
        x_0 = [];
    end
    
     if isempty(x_0)
        free_ci = 1;
     end
    
    if nargin < 11
        param_fixe = [];
    end
    
     gradient_use = 1;
    if isempty(der_vect_field_theta) || isempty(der_vect_field_state_theta)  || isempty(der_vect_field_state_state)
        gradient_use = 0;
    end
    
    
    function res_t = champ_vect_state_param(t,vx,vp,param)
        res_t = vect_field(t,vx,param,param_fixe)+ invU*vp/2;
    end


     function res_t = champ_vect_adjoint_param(t,vx,vp,param)
       y_hat = bs.calcul_prediction(t,0);
       y_hat = reshape(y_hat,dim_obs,1);

       res_t = 2*(mat_Obs')*mat_Obs*vx-(der_vect_field(t,vx,param,param_fixe)')*vp-2*(mat_Obs')*y_hat;
     end
    

   


      function [res_f res_grad]= func_eco_multi_param(param)
         % param

         nb_param_ci = length(param);
          
         if free_ci ==1
             nb_param = nb_param_ci - dim_syst;
             nv_x_0 = param((nb_param+1):end);
             param_tot = param(1:nb_param);
             nb_ci = dim_syst;
         else
             nb_param = nb_param_ci;
             nv_x_0 = x_0;
             param_tot = param;
             nb_ci = 0;
         end
          
         %nv_x_0
         nb_param = length(param_tot);
        
        
          function dy = equ_b(t,y)
                val_x = reshape(y(1:dim_syst),dim_syst,1); 
                val_p = reshape(y((dim_syst+1):2*dim_syst),dim_syst,1); 

                dx = champ_vect_state_param(t,val_x,val_p,param_tot);
                dp = champ_vect_adjoint_param(t,val_x,val_p,param_tot);

                dx_v = reshape(dx,dim_syst,1);
                dp_v = reshape(dp,dim_syst,1);

                dy = [dx_v' dp_v']';

          end
       
            function res = bcfun(sol_T_0,sol_T_end)
                res = zeros(2*dim_syst,1);
                res(1:dim_syst,1) = sol_T_0(1:dim_syst)-reshape(nv_x_0,dim_syst,1);
                res((dim_syst+1):2*dim_syst,1) = sol_T_end((dim_syst+1):2*dim_syst);
            end
            
            function dy = equ_diff_ini(t,y)
                dy = vect_field(t,y,param_tot,param_fixe);
            end
            sol_ini = ode45(@equ_diff_ini,Int_ode_solve,nv_x_0);   
         
            function sol_t_ini = func_sol_ini(t)
                res_ini_t = deval(sol_ini,t);
                sol_t_ini = zeros(2*dim_syst,1);
                sol_t_ini(1:dim_syst,1) = res_ini_t; 
            end

         solinit = bvpinit(Int_ode_solve,@func_sol_ini);    

         sol_equ =  bvp5c(@equ_b,@bcfun,solinit);

         function dy = equ_sens_b(t,y)
                val_x_sens = y(1:dim_syst*nb_param);
                val_p_sens = y((dim_syst*nb_param+1):(2*dim_syst*nb_param));
                
                val_x_sens = reshape(val_x_sens,dim_syst,nb_param);
                val_p_sens = reshape(val_p_sens,dim_syst,nb_param);
                
                val_x_p =  deval(sol_equ,t);
                val_x = reshape(val_x_p(1:dim_syst),dim_syst,1);
                val_p = reshape(val_x_p((dim_syst+1):2*dim_syst),dim_syst,1);
                
                df_state_theta = der_vect_field_state_theta(t,val_x,param_tot,param_fixe);
                df_state_state = der_vect_field_state_state(t,val_x,param_tot,param_fixe);
                df_theta  = der_vect_field_theta(t,val_x,param_tot,param_fixe);
                df_state  = der_vect_field(t,val_x,param_tot,param_fixe);
                
                
                part_sens_p_der_theta_inter = zeros(dim_syst,nb_param);
                part_sens_p_der_x_inter = zeros(dim_syst);
                
                for nd = 1 : dim_syst
                     part_sens_p_der_theta_inter =  part_sens_p_der_theta_inter +df_state_theta(:,:,nd)*val_p(nd);
                     part_sens_p_der_x_inter = part_sens_p_der_x_inter + df_state_state(:,:,nd)*val_p(nd);
                end
                
                part_sens_p_der_theta = -part_sens_p_der_theta_inter;
                part_sens_p_der_x = - part_sens_p_der_x_inter +2*(mat_Obs')*mat_Obs;
                
                d_val_x_sens = df_theta+(invU/2)*val_p_sens+df_state*val_x_sens;
                d_val_p_sens = part_sens_p_der_theta-(df_state')*val_p_sens + part_sens_p_der_x*val_x_sens;
                
                dy = [reshape(d_val_x_sens,1,dim_syst*nb_param) reshape(d_val_p_sens,1,dim_syst*nb_param)]';
         end
          
         
         function res = bcfun_sens(sol_sens_T_0,sol_sens_T_end)
                res = zeros(2*dim_syst*nb_param,1);
                res(1:dim_syst*nb_param,1) = sol_sens_T_0(1:(dim_syst*nb_param));
                res((dim_syst*nb_param+1):2*dim_syst*nb_param,1) = sol_sens_T_end((nb_param*dim_syst+1):(2*dim_syst*nb_param));
         end
         
         
         function dy = equ_sens_b_ini(t,y)
                val_x_sens = y(1:dim_syst*nb_param);
                val_x_sens = reshape(val_x_sens,dim_syst,nb_param);
                
                val_x =  deval(sol_ini,t);
                val_x = reshape(val_x,dim_syst,1);
                df_theta  = der_vect_field_theta(t,val_x,param_tot,param_fixe);
                df_state  = der_vect_field(t,val_x,param_tot,param_fixe);
                                              
                d_val_x_sens = df_theta+df_state*val_x_sens;
                
                dy = reshape(d_val_x_sens,1,dim_syst*nb_param)';
         end
         
         
           function sol_t_ini = func_sol_ini_sens(t)
                sens_x_p_nul = deval(sol_equ_sens_ini,t);
                sol_t_ini = zeros(2*dim_syst*nb_param,1);
                sol_t_ini(1:dim_syst*nb_param) = reshape(sens_x_p_nul,dim_syst*nb_param,1); 
           end

         
         if gradient_use ==1
             sol_equ_sens_ini = ode45(@equ_sens_b_ini,Int_ode_solve,zeros(1,dim_syst*nb_param));   
             solinit_sens = bvpinit(Int_ode_solve,@func_sol_ini_sens);    
             sol_equ_sens =  bvp5c(@equ_sens_b,@bcfun_sens,solinit_sens);
         end
         
          function dy = equ_sens_x0_pT(t,y)
                val_x_sens = y(1:dim_syst*dim_syst);
                val_p_sens = y((dim_syst*dim_syst+1):(2*dim_syst*dim_syst));
                
                val_x_sens = reshape(val_x_sens,dim_syst,dim_syst);
                val_p_sens = reshape(val_p_sens,dim_syst,dim_syst);
                
                val_x_p =  deval(sol_equ,t);
                val_x = reshape(val_x_p(1:dim_syst),dim_syst,1);
                val_p = reshape(val_x_p((dim_syst+1):2*dim_syst),dim_syst,1);
                
                df_state_state = der_vect_field_state_state(t,val_x,param_tot,param_fixe);
                df_state  = der_vect_field(t,val_x,param_tot,param_fixe);

                part_sens_p_der_x_inter = zeros(dim_syst,dim_syst);
                
                for nd = 1 : dim_syst
                     part_sens_p_der_x_inter = part_sens_p_der_x_inter + df_state_state(:,:,nd)*val_p(nd);
                end
                
                part_sens_p_der_x = - part_sens_p_der_x_inter +2*(mat_Obs')*mat_Obs;
                
                d_val_x_sens = (invU/2)*val_p_sens+df_state*val_x_sens;
                d_val_p_sens = -(df_state')*val_p_sens + part_sens_p_der_x*val_x_sens;
                
                dy = [reshape(d_val_x_sens,1,dim_syst*dim_syst) reshape(d_val_p_sens,1,dim_syst*dim_syst)]';
          end
         
             function res = bcfun_sens_x0_pT(sol_sens_x0_T_0,sol_sens_T_end)
                res = zeros(2*dim_syst*dim_syst,1);
                res(1:dim_syst*dim_syst,1) = sol_sens_x0_T_0(1:(dim_syst*dim_syst)) - reshape(eye(dim_syst,dim_syst),dim_syst*dim_syst,1);
                res((dim_syst*dim_syst+1):2*dim_syst*dim_syst,1) = sol_sens_T_end((dim_syst*dim_syst+1):(2*dim_syst*dim_syst));
             end
             
             
                 
         function dy = equ_sens_b_ini_x0_pT(t,y)
                val_x_sens = y(1:dim_syst*dim_syst);
                val_x_sens = reshape(val_x_sens,dim_syst,dim_syst);
                
                val_x =  bs.calcul_prediction(t,0);
                val_x = reshape(val_x,dim_syst,1);
                df_state  = der_vect_field(t,val_x,param_tot,param_fixe);
                                              
                d_val_x_sens = df_state*val_x_sens;
                
                dy = reshape(d_val_x_sens,1,dim_syst*dim_syst)';
         end
         
                 
           function sol_t_ini = func_sol_ini_sens_x0_pT(t)
                sens_x_p_nul = deval(sol_equ_sens_ini_x0_pT,t);
                sol_t_ini = zeros(2*dim_syst*dim_syst,1);
                sol_t_ini(1:dim_syst*dim_syst) = reshape(sens_x_p_nul,dim_syst*dim_syst,1); 
           end

            
           if (nb_ci >0) && (gradient_use ==1)
             sol_equ_sens_ini_x0_pT = ode45(@equ_sens_b_ini_x0_pT,Int_ode_solve,reshape(eye(dim_syst,dim_syst),1,dim_syst*dim_syst));
             solinit_sens_x0_pT = bvpinit(Int_ode_solve,@func_sol_ini_sens_x0_pT);    
             sol_equ_sens_x0_pT =  bvp4c(@equ_sens_x0_pT,@bcfun_sens_x0_pT,solinit_sens_x0_pT);
           end
 
         
         function [res_f_t res_grad_t]= func_inner_integ(t)
            
            %res_f_t = zeros(1+(nb_param+nb_ci)*gradient_use,1);
            res_f_t = zeros(dim_syst+dim_obs,1);
            y_hat = bs.calcul_prediction(t,0);
            y_hat = reshape(y_hat,dim_obs,1);
            val_x_p = deval(sol_equ,t);
            
            val_x = reshape(val_x_p(1:dim_syst),dim_syst,1);
            val_p = reshape(val_x_p((dim_syst+1):2*dim_syst),dim_syst,1);
            
            val_u = invU*val_p/2;
            
            res_f_t(1:dim_obs) = (mat_Obs*val_x - y_hat);
            res_f_t((dim_obs+1):(dim_obs+dim_syst)) = val_u*sqrt(mat_U(1,1));
            
           if gradient_use == 1
                val_x_sens_tot = zeros(dim_syst,nb_param+nb_ci);
                val_p_sens_tot = zeros(dim_syst,nb_param+nb_ci);
                res_grad_t = zeros(dim_syst+dim_obs,nb_param+nb_ci);

                val_x_p_sens = deval(sol_equ_sens,t);
                val_x_sens = val_x_p_sens(1:dim_syst*nb_param);
                val_p_sens = val_x_p_sens((dim_syst*nb_param+1):(2*dim_syst*nb_param));
                val_x_sens = reshape(val_x_sens,dim_syst,nb_param);
                val_p_sens = reshape(val_p_sens,dim_syst,nb_param);

                val_x_sens_tot(:,1:nb_param) = val_x_sens;
                val_p_sens_tot(:,1:nb_param) = val_p_sens;

                if nb_ci > 0
                    val_x_p_sens_xT = deval(sol_equ_sens_x0_pT,t);
                    val_x_sens_xT = val_x_p_sens_xT(1:dim_syst*dim_syst);
                    val_p_sens_xT = val_x_p_sens((dim_syst*dim_syst+1):(2*dim_syst*dim_syst));
                    val_x_sens_xT = reshape(val_x_sens_xT,dim_syst,dim_syst);
                    val_p_sens_xT = reshape(val_p_sens_xT,dim_syst,dim_syst);

                    val_x_sens_tot(:,(nb_param+1):(nb_param+dim_syst)) = val_x_sens_xT;
                    val_p_sens_tot(:,(nb_param+1):(nb_param+dim_syst)) = val_p_sens_xT;
                end

                res_grad_t(1:dim_obs,:) = mat_Obs*val_x_sens_tot;
                
                val_sens_u_tot = invU*val_p_sens_tot/2;
                
                res_grad_t((dim_obs+1):(dim_obs+dim_syst),:) =  val_sens_u_tot*sqrt(mat_U(1,1));
           end
         end
        
        pas_solve_sq = 100; 
        Int_ode_solve_sq = Times(1):((Times(end)-Times(1))/pas_solve_sq):Times(end);          
         [res_f res_grad] = arrayfun(@func_inner_integ,Int_ode_solve_sq,'UniformOutput',false);
         
         res_f = cell2mat(res_f');
         res_grad = cell2mat(res_grad');
        
    
      end
  
  

        nb_fun_eval = 1000;
        max_iter = 150;
        options = optimset('Display','iter','TolFun',1e-7,'TolX',1e-7,'GradObj','on','MaxFunEvals',nb_fun_eval,'MaxIter',max_iter,'Algorithm','levenberg-marquardt');
        [est_param,fval]= lsqnonlin(@func_eco_multi_param,param_ini,[],[],options);


end