function [est_param,fval,bs_u] = est_param_pontryagin_bvp4c_adjoint_method(bs,vect_field,der_vect_field,der_vect_field_theta,der_vect_field_state_theta,der_vect_field_state_state,mat_Obs,mat_U,param_ini,x_0_known,mat_b,type_solver,bvp_solv,param_fixe)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%DESCRIPTION: % Compute the Pontryagin Maximum estimator for a given value of lambda
%
%USAGE:    [est_param,fval,bs_u] =  est_param_pontryagin_bvp4c_adjoint_method(bs,vect_field,der_vect_field,der_vect_field_theta,der_vect_field_state_theta,der_vect_field_state_state,mat_Obs,mat_U,param_ini,x_0_known,type_solver,bvp_solv,param_fixe)
%
%INPUT :
%bs               :(bSplines  object) , embed the nonparametric estimators
%vect_field       :(function handler) embed the function corresponding to the vector field with signature: 
%                  res_t =  vect_field(t,vx,param,param_fixe)
%                           t : time
%                           vx : state value at time t
%                           param : varying parameter value for the plotted graph
%                           param_fixe : other parameter value
%                           res_t : (vector of size dim x 1) vector field  value at time t
%der_vect_field       :(function handler) embed the function corresponding to the vector field derivative form x.rt state with signature: 
%                  res_t = der_vect_field(t,vx,param,param_fixe)
%                           t : time
%                           vx : state value at time t
%                           param : varying parameter value for the plotted graph
%                           param_fixe : other parameter value
%                           res_t : (matrix of size dim x dim) matrix der_f_i/x_j: vector field derivative form x.rt state value at time t
%der_vect_field_theta       :(function handler) embed the function corresponding to the vector field derivative form x.rt parameter with signature: 
%                  res_t = der_vect_field(t,vx,param,param_fixe)
%                           t : time
%                           vx : state value at time t
%                           param : varying parameter value for the plotted graph
%                           param_fixe : other parameter value
%                           res_t : (matrix of size dim x nb param) matrix der_f_i/theta_j: vector field derivative form x.rt parameter value at time t
%der_vect_field_state_theta       :(function handler) embed the function corresponding to the vector field second derivative form x.rt state and parameter with signature: 
%                  res_t = der_sec_vect_field(t,vx,param,param_fixe)
%                           t : time
%                           vx : state value at time t
%                           param : varying parameter value for the plotted graph
%                           param_fixe : other parameter value
%                           res_t : (hypermatrix of size dim x nb param x dim) matrix der_sec_f_i/x_j_theta_h: vector field second derivative form x.rt state and parameter value at time t
%der_vect_field_state_state       :(function handler) embed the function corresponding to the vector field second derivative form x.rt state with signature: 
%                  res_t = der_sec_vect_field(t,vx,param,param_fixe)
%                           t : time
%                           vx : state value at time t
%                           param : varying parameter value for the plotted graph
%                           param_fixe : other parameter value
%                           res_t : (hypermatrix of size dim x dim x dim) matrix der_sec_f_i/x_j_x_h: vector field second derivative form x.rt state value at time t
%mat_Obs                  : (matrix of size dim_obs x dim): observation matrix C
%mat_U                    : (matrix of size dim x dim): ponderation matrix U for control
%param_ini                : (matrix of size nb param) initial guess for the optimization algorithm
%OPTIONAL
%x_0_known                :(vector ) initial condition if known (default empty [] )
%bvp_solv                 : (integer) indicate the used mixed boundary ode solver
%                         1 (default) bvp4c
%                         0           bvp5c
%type_solv                :(integer) indicate the used optimization matlab function
%                          1 (default) fminunc
%                          0 fmincon
%param_fixe               :(vector) fixed parameter for the model (default empty ([]) )
%OUTPUT:
%est_param                :(vector) estimated value
%res_fval                 :(matrix or vector) economical function value 
%bs_u                     :(bSplines object) estimated optimal control              
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% 

    Obs = bs.getObservation();
    T_c = Obs.getTimes();
    Times = T_c{1};
    invU = inv(mat_U);
    dim_syst = length(mat_Obs(1,:));
    dim_obs = length(mat_Obs(:,1));
    
    NMax_val = 300;

    pas_solve =400;
    Int_ode_solve = Times(1):((Times(end)-Times(1))/pas_solve):Times(end);          
    
    free_ci = 0;
    if nargin < 10
        x_0_known = [];
    end
    
     if isempty(x_0_known)
        free_ci = 1;
     else
         if length(x_0_known) < dim_syst
                 free_ci = 1;
         end
     end
     
     
     nb_param_ci = length(param_ini);
     if free_ci == 1
             nb_ci_known = length(x_0_known);
             nb_ci = dim_syst-nb_ci_known;             
             nb_param = nb_param_ci - nb_ci;
     else
             nb_param = nb_param_ci;
             nb_ci = 0;
     end
    
     
    if nargin < 11
        mat_b = eye(dim_syst);
    end
    
    dim_control =  length(mat_b(1,:));
    mat_weight_p = mat_b*invU *(mat_b');
    
    if nargin < 14
        param_fixe = [];
    end
    
    if nargin < 13
        bvp_solv = 1;
    end
    
    if nargin < 12
        type_solver = 1;
    end
    
     gradient_use = 1;
    if isempty(der_vect_field_theta) || isempty(der_vect_field_state_theta)  || isempty(der_vect_field_state_state)
        gradient_use = 0;
    end
    
    
    function res_t = champ_vect_state_param(t,vx,vp,param)
        res_t = vect_field(t,vx,param,param_fixe)+ mat_weight_p*vp/2;
    end


     function res_t = champ_vect_adjoint_param(t,vx,vp,param)
       y_hat = bs.calcul_prediction(t,0);
       y_hat = reshape(y_hat,dim_obs,1);

       res_t = 2*(mat_Obs')*mat_Obs*vx-(der_vect_field(t,vx,param,param_fixe)')*vp-2*(mat_Obs')*y_hat;
     end
 
 
    function [res_in res_eq] = mycon(param)
        res_in = 0;
        res_eq = 0;
    end
 
     function [res_val res_grad]= func_eco_multi_param(param)
         
         curr_compute_grad = 0;
         if nargout > 1
             curr_compute_grad = 1;
         end

          %param
         if free_ci ==1
             nv_x_0_var = param((nb_param+1):end);
             nv_x_0_var = reshape(nv_x_0_var,1,length(nv_x_0_var));
             nv_x_0 = [x_0_known nv_x_0_var];
             param_tot = param(1:nb_param);
             param_tot = reshape(param_tot,1,length(param_tot));

         else
             nv_x_0 = x_0_known;
             param_tot = param;
         end
          
         %nv_x_0
         nb_param = length(param_tot);   
         

          function dy = equ_b(t,y)
                val_x = reshape(y(1:dim_syst),dim_syst,1); 
                val_p = reshape(y((dim_syst+1):2*dim_syst),dim_syst,1); 

                dx = champ_vect_state_param(t,val_x,val_p,param_tot);
                dp = champ_vect_adjoint_param(t,val_x,val_p,param_tot);

                dx_v = reshape(dx,dim_syst,1);
                dp_v = reshape(dp,dim_syst,1);

                dy = [dx_v' dp_v']';

          end
       
            function res = bcfun(sol_T_0,sol_T_end)
                res = zeros(2*dim_syst,1);
                res(1:dim_syst,1) = sol_T_0(1:dim_syst)-reshape(nv_x_0,dim_syst,1);
                res((dim_syst+1):2*dim_syst,1) = sol_T_end((dim_syst+1):2*dim_syst);
            end
            
            function dy = equ_b_ini(t,y)
                val_x = reshape(y,dim_syst,1);        
                dy = vect_field(t,val_x,param_tot,param_fixe);
             end
         
         
           function sol_t_ini = func_sol_ini(t)
                sens_x_p = deval(sol_ini,t);
                sol_t_ini = zeros(2*dim_syst,1);
                sol_t_ini(1:dim_syst) = reshape(sens_x_p,dim_syst,1); 
           end

         

        if (gradient_use ==0 )  || (curr_compute_grad == 0 ) 
         sol_ini = ode45(@equ_b_ini,Int_ode_solve,reshape(nv_x_0,dim_syst,1));      
         solinit = bvpinit(Int_ode_solve,@func_sol_ini);   
         options_ode = bvpset('NMax',NMax_val,'RelTol',1e-06,'AbsTol',1e-08);
         
         if bvp_solv == 1
             sol_equ =  bvp4c(@equ_b,@bcfun,solinit,options_ode);
         else
             sol_equ =  bvp5c(@equ_b,@bcfun,solinit,options_ode);
         end
        end
        
        ind_fin_p = 2*dim_syst;
        ind_fin_adjoint = ind_fin_p+2*dim_syst;


         function dy = equ_adjoint_tot_b(t,y)
                val_x = reshape(y(1:dim_syst),dim_syst,1); 
                val_p = reshape(y((dim_syst+1): ind_fin_p),dim_syst,1); 

                val_adjoint = y((ind_fin_p+1):ind_fin_adjoint);
                val_adjoint = reshape(val_adjoint,2*dim_syst,1);
                
                val_curve = bs.calcul_prediction(t,0);
                val_curve = reshape(val_curve,dim_obs,1);
                
                
                dx = champ_vect_state_param(t,val_x,val_p,param_tot);
                dp = champ_vect_adjoint_param(t,val_x,val_p,param_tot);

                dx_v = reshape(dx,dim_syst,1);
                dp_v = reshape(dp,dim_syst,1);

       
                df_state_state = der_vect_field_state_state(t,val_x,param_tot,param_fixe);
                df_state  = der_vect_field(t,val_x,param_tot,param_fixe);
                
                der_F_state = zeros(2*dim_syst);
                der_l_state = zeros(2*dim_syst,1);
                
                part_sens_p_der_x_inter = zeros(dim_syst);
                
                for nd = 1 : dim_syst
                     part_sens_p_der_x_inter = part_sens_p_der_x_inter + df_state_state(:,:,nd)*val_p(nd);
                end

                part_sens_p_der_x = - part_sens_p_der_x_inter +2*(mat_Obs')*mat_Obs;
                
                der_F_state(1:dim_syst,1:dim_syst) = df_state;
                der_F_state(1:dim_syst,(dim_syst+1):2*dim_syst) = mat_weight_p/2;
                der_F_state((dim_syst+1):2*dim_syst,1:dim_syst) = part_sens_p_der_x;
                der_F_state((dim_syst+1):2*dim_syst,(dim_syst+1):2*dim_syst) = -df_state';
                
                der_l_state(1:dim_syst) = 2*(mat_Obs')*(mat_Obs*val_x -val_curve);
                der_l_state((dim_syst+1):2*dim_syst) = mat_weight_p*val_p/2;
                
                d_val_adjoint = der_l_state - (der_F_state')*val_adjoint;
                
                dy = [dx_v' dp_v' d_val_adjoint']';
                
                
         end
          
         
         function res = bcfun_adjoint_tot(sol_adjoint_tot_T_0,sol_adjoint_tot_T_end)
             
                res = zeros(4*dim_syst,1);
             
                res(1:dim_syst,1) = sol_adjoint_tot_T_0(1:dim_syst)-reshape(nv_x_0,dim_syst,1);
                res((dim_syst+1):ind_fin_p,1) = sol_adjoint_tot_T_end((dim_syst+1):ind_fin_p);
                res((ind_fin_p+1):(ind_fin_p+dim_syst),1) = sol_adjoint_tot_T_end((ind_fin_p+1):(ind_fin_p+dim_syst));
                res((ind_fin_p+dim_syst+1):ind_fin_adjoint,1) = sol_adjoint_tot_T_0((ind_fin_p+dim_syst+1):ind_fin_adjoint);

         end
         
         

         
           function sol_t_ini = func_sol_ini_adjoint(t)             

                val_x_ini = deval(sol_ini,t);
                sol_t_ini = zeros(4*dim_syst,1);
                sol_t_ini(1:dim_syst,1) = val_x_ini;
           end

         
         if (gradient_use ==1) && (curr_compute_grad ==1)          
             sol_ini = ode45(@equ_b_ini,Int_ode_solve,reshape(nv_x_0,dim_syst,1));      
             solinit_adjoint = bvpinit(Int_ode_solve,@func_sol_ini_adjoint);  
             options_ode_gd = bvpset('NMax',NMax_val,'RelTol',1e-06,'AbsTol',1e-08);
             
             if bvp_solv == 1
                 sol_equ_adjoint =  bvp4c(@equ_adjoint_tot_b,@bcfun_adjoint_tot,solinit_adjoint,options_ode_gd);
             else
                 sol_equ_adjoint =  bvp5c(@equ_adjoint_tot_b,@bcfun_adjoint_tot,solinit_adjoint,options_ode_gd);
             end
         end
         
 
         
         function res_t = func_inner_integ(t)
             
                res_t = zeros(1,1+nb_param*gradient_use);
                y_hat = bs.calcul_prediction(t,0);
                y_hat = reshape(y_hat,dim_obs,1);
            
          if (gradient_use == 1)   && (curr_compute_grad ==1)         
                val_x_p_adjoint = deval(sol_equ_adjoint,t);
                val_x = val_x_p_adjoint(1:dim_syst);
                val_p = val_x_p_adjoint((dim_syst+1):ind_fin_p);
                val_adjoint = val_x_p_adjoint((ind_fin_p+1):(ind_fin_adjoint));
                val_adjoint = reshape(val_adjoint,1,2*dim_syst);
                
                
                df_theta  = der_vect_field_theta(t,val_x,param_tot,param_fixe);
                df_state_theta = der_vect_field_state_theta(t,val_x,param_tot,param_fixe);
                
                
                Der_F_param = zeros(2*dim_syst,nb_param);
                part_sens_p_der_theta_inter = zeros(dim_syst,nb_param);

                for nd = 1 : dim_syst
                     part_sens_p_der_theta_inter =  part_sens_p_der_theta_inter +df_state_theta(:,:,nd)*val_p(nd);
                end
                
                part_sens_p_der_theta = -part_sens_p_der_theta_inter;
                
                Der_F_param(1:dim_syst,:) = df_theta;
                Der_F_param((dim_syst+1):2*dim_syst,:) = part_sens_p_der_theta;
            
            
            
            res_t(1) = sum((mat_Obs*val_x - y_hat).^2) + (val_p')*mat_weight_p*val_p/4;           
            res_t(2:end) = -val_adjoint*Der_F_param;
          else

                val_x_p = deval(sol_equ,t);

                val_x = reshape(val_x_p(1:dim_syst),dim_syst,1);
                val_p = reshape(val_x_p((dim_syst+1):2*dim_syst),dim_syst,1);

                res_t(1) = sum((mat_Obs*val_x - y_hat).^2) + (val_p')*mat_weight_p*val_p/4;  
          end
        end
        
        res_integration = quadv(@func_inner_integ,Times(1),Times(end));      
        res_val = res_integration(1);
        
        if (gradient_use == 1)   && (curr_compute_grad ==1)     
           res_grad = zeros(nb_param+nb_ci,1); 
           res_grad(1:nb_param) = res_integration(2:end);
           if free_ci ==1
               val_x_p_val_ini = deval(sol_equ_adjoint,Times(1));
               val_adjoint_p1_ini = val_x_p_val_ini((ind_fin_p+1):(ind_fin_p+dim_syst));
               res_grad((nb_param+1):(nb_param+nb_ci)) = -val_adjoint_p1_ini((nb_ci_known+1): dim_syst);
           end
        end
        
      end
  
  
  
       if gradient_use == 1
           if  type_solver == 1;   
              max_iter = 300;
              nb_fun_eval = 500; 
            options = optimset('TolFun',1e-7,'TolX',1e-7,'Display','iter','GradObj','on','MaxFunEvals',nb_fun_eval,'MaxIter',max_iter,'Hessian','off');
            [est_param,fval]= fminunc(@func_eco_multi_param,param_ini,options);
            
          
           else     
                 lb = zeros(length(param_ini),1);
                 ub = zeros(length(param_ini),1);
                 ub(1:length(param_ini)) = +Inf;
                 lb(1:length(param_ini)) = -Inf;
                 max_iter = 10000;
                 nb_fun_eval = 10000; 
                 options = optimset('TolFun',1e-7,'TolX',1e-7,'Display','iter','GradObj','on','MaxFunEvals',nb_fun_eval,'MaxIter',max_iter);
                
                   if type_solver == 3
                        'fmincon Mat_A'
                       Mat_A = -eye(length(param_ini));
                       Mat_A(4,4) = 2;
                       vect_b = zeros(length(param_ini),1);
                       [est_param,fval] =  fmincon(@func_eco_multi_param,param_ini,Mat_A,vect_b,[],[],lb,ub,@mycon,options);
                   else
                        if type_solver == 2 
                          [est_param,fval] =  fmincon(@func_eco_multi_param,param_ini,[],[],[],[],lb,ub,@mycon,options);
                        else
                          [est_param,fval] = gradient_descent_backtracking(@func_eco_multi_param,param_ini,10^-5,10^-5);
                          %nb_fun_eval = 100000; 
                        end
                    end
          end

      else
       %[est_param,fval]= fminsearch(@func_eco_multi_param,param_ini);
              max_iter = 150;
              nb_fun_eval = 500;  
              options = optimset('TolFun',1e-7,'TolX',1e-7,'Display','iter','GradObj','off','MaxFunEvals',nb_fun_eval,'MaxIter',max_iter,'Hessian','off');
            [est_param,fval]= fminunc(@func_eco_multi_param,param_ini,options);
       end

       if nargout > 2
      
             if free_ci == 1
                est_param_theta = est_param(1:nb_param);
                est_param_x_0 = est_param((nb_param+1):end); 
                x_0_tot_est = [x_0_known est_param_x_0];
             else
                est_param_theta = est_param;
                est_param_x_0 = x_0_known; 
                x_0_tot_est = est_param_x_0;
             end
            try
                bs_u = est_control_pontryagin(bs,vect_field,der_vect_field,mat_Obs,mat_U,est_param_theta,x_0_tot_est,mat_b,param_fixe);
            catch
                    Obs_bsu = Observation(Times,zeros(length(Times),dim_control)); 
                    bs_u = bSplines(Obs_bsu);
            end
           
       end

end